/*-------------------------------------------------------------------*/
/* MAPSEG.C      Pasi Frnti                                         */
/*                                                                   */
/* Color separation of map images to binary layers                   */
/* Input:  PPM color image                                           */
/* Output: Series of PBM binary images                               */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "MAPSEG"
#define VersionNumber  "Version 0.01"
#define LastUpdated    "12.4.01"

/*-------------------------------------------------------------------*/

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#define  FACTFILE  "mapseg.fac"

#include "bintree.h"
#include "cb.h"
#include "file.h"
#include "image.h"
#include "parametr.c"
#include "pgm.h"
#include "memctrl.h"
#include "owntypes.h"


/* ========================= Time functions ============================= */


static long SetWatch(long* start)
{
  time_t tmp;
  return( *start = (long)time(&tmp) );
}


/* ----------------------------------------------------------------- */


static void PrintTime(long start)
{
  time_t  tmp;
  int     hours, minutes, seconds;
  long    elapsed;

  elapsed  = (long)time(&tmp) - start;
  hours    = (int)(elapsed / 3600);
  elapsed  = (int)(elapsed % 3600);
  minutes  = (int)(elapsed / 60);
  seconds  = (int)(elapsed % 60);

  switch( Value(QuietLevel) )
    {
    case 0:  break;
    case 1:  printf(" %i:%02i:%02i", hours, minutes, seconds);
             break;
    default: printf("Time: %i:%02i:%02i\n", hours, minutes, seconds);
             break;
    }
}


/*==========================  P R I N T  ============================*/


void PrintInfo(void)
{
  printf("%s     %s                (last modified %s)\n",ProgName,VersionNumber,LastUpdated);
  printf("Bit plane separation.\n");
  printf("Usage: %s inputfile outputfile [%coptions]\n", ProgName, OPTION_SYMBOL);
  printf("\n  Options:\n");
  PrintOptions();
  printf("\n%s     %s\n",ProgName,VersionNumber);
}


/*-------------------------------------------------------------------*/


static void PrintOperatingInfo(char* InputFileName, char* OutputFileName,
                               IMAGE* InputImage)
{
  if( Value(QuietLevel) >= 1 )
    {
    printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
    printf("Input file:       %s  (%ix%i)\n", InputFileName,
            InputImage->ImageSizeX, InputImage->ImageSizeY);
    printf("Output files:     %s1.PBM .. %sN.PBM \n", OutputFileName, OutputFileName);
    printf("\n");
    PrintSelectedOptions();
    printf("\n");
    }
}


/*-------------------------------------------------------------------*/


static void PrintProgress(int y, int total)
{
  if( Value(QuietLevel) >= 2 )
    {
    printf("%5.1f %% done.\r", (float)(y + 1) * 100 / total);
    fflush(stdout);
    }
}


/* =========================  P R O C E S S  ======================= */


void InitializeTrainingSet(TRAININGSET* TS, int size, int maxvalue)
{
  char GenMethod[MaxGenMethodLength];
  
  strcpy(GenMethod, ProgName);
  strcat(GenMethod, " ");
  strcat(GenMethod, VersionNumber);
  CreateNewTrainingSet(TS, size, 3, 1, 1, 0, maxvalue, 0, GenMethod);
  BookSize(TS) = 0;
}


/*-------------------------------------------------------------------*/


int CompareNodes(void* a, void* b, void* info)
{
  return ( CompareVectors( ((BOOKNODE*) a)->vector,
                           ((BOOKNODE*) b)->vector, 3 ));
}    


/* ----------------------------------------------------------------- */


static void InsertVectorToBintree(BINTREE* Tree, VECTORTYPE v)
{
  BOOKNODE* node;
  void*     found;

  node  = allocate(sizeof(BOOKNODE));
  *node = CreateEmptyNode(3);
  CopyVector(v, node->vector, 3);
  node->freq = 1;

  found = InsertToBintreeNoDuplicates( Tree, (void*) node, NULL);
  if (found)
    {
    ((BOOKNODE*) found)->freq++;
    FreeNode(*node);
    deallocate(node);
    }
}    


/*-------------------------------------------------------------------*/


static void ReadPixelsToBinaryTree(IMAGE* Image, BINTREE* Tree)
{
  VECTORTYPE  v;
  int x, y;
  int Ystep=1;
  int Xstep=1;

  if(Value(QuietLevel)>=2) printf("Reading image into binary tree.\n");
  v = CreateEmptyVector(3);

  for( y=1; y<=Image->ImageSizeY; y+=Ystep )
     {
     ScrollBuffer(Image);
	 for( x=1; x<=Image->ImageSizeX; x+=Xstep )
        {
        v[0] = GetImageRed(  Image,x,0);
        v[1] = GetImageGreen(Image,x,0);
        v[2] = GetImageBlue( Image,x,0);
        InsertVectorToBintree(Tree, v);
        /* printf("(%03i,%03i,%03i)  ", v[0], v[1], v[2]); */
        }
     PrintProgress(y, Image->ImageSizeY);
     }

  if((Value(QuietLevel)) >= 3)
    {
    printf("Training set contains now %u cells.\n", BintreeSize(Tree));
    }

  FreeVector(v);
}


/*-------------------------------------------------------------------*/


static void InsertNodeToTrainingSet(TRAININGSET* TS, BOOKNODE* node)
{
  CopyNode(node, &Node(TS, BookSize(TS)), 3);
  TotalFreq(TS) += node->freq;
  BookSize(TS)++;
}    


/*-------------------------------------------------------------------*/


static void ConvertTreeToTrainingSet(TRAININGSET* TS, BINTREE* Tree)
{
  STACK     S;
  BOOKNODE* node;
  
  if(Value(QuietLevel)>=2) printf("Converting binary tree to training set.\n");
  if(Value(QuietLevel)>=2) printf("Tree contains following colors:\n");

  InitInOrderBintree(Tree, &S);
  
  node = (BOOKNODE*) InOrderBintree(&S);
  while (node != NULL)
    {
    InsertNodeToTrainingSet(TS, node);
    PrintVector(node->vector,3,1);
    FreeNode(*node);
    deallocate(node);
    node = (BOOKNODE*) InOrderBintree(&S);
    }
}


/*-------------------------------------------------------------------*/


static void GenerateCodebook(TRAININGSET* TS, CODEBOOK* C, PARTITIONING* P)
{
  if(Value(QuietLevel)>=2) printf("Generating codebook.\n");

  CopyCodebook(TS, C);
}


/*-------------------------------------------------------------------*/


static llong RGBdistance(VECTORTYPE v, RGBPIXEL color)
{
  llong diff = 0;
  llong dist = 0;

  diff	= (llong)v[0] - (llong)color.Red;
  dist += diff * diff;
  diff	= (llong)v[1] - (llong)color.Green;
  dist += diff * diff;
  diff	= (llong)v[2] - (llong)color.Blue;
  dist += diff * diff;

  return(dist);
}


/*-------------------------------------------------------------------*/


static int FindNearestColor(CODEBOOK* C, RGBPIXEL color)
{
  llong  d;
  llong  mindist  = MAXLLONG;
  int    minindex = 0;
  int    i;

  for( i=0; i<BookSize(C); i++ )
     {
     d = RGBdistance(Vector(C,i), color);
     if( d < mindist )
       {
       mindist  = d;
       minindex = i;
       if( d == 0 ) return(minindex);
       }
    }
  return(minindex);
}


/*-------------------------------------------------------------------*/
#define  MAX_LAYERS   8


static void PerformColorSeparation(IMAGE*    InputImage,
								   char*	 PlainName,
                                   CODEBOOK* C)
{
  IMAGE     OutputImage[MAX_LAYERS+1];
  IMAGE     QuantizedImage;
  FILE*     OutputFile[MAX_LAYERS+1];
  FILE*     QuantizedFile = NULL;
  char      OutputFileName[64];
  RGBPIXEL  color, color2;
  int       index,i,x,y;

  if(Value(QuietLevel)>=2) printf("Perform color separation.\n");

  /* Prepare input and output images */
  for(i=1; i<=BookSize(C); i++)
     {
     /* Generate name for output files */
     PickFileName(PlainName, OutputFileName);
     sprintf(OutputFileName, "%s-%i", PlainName, i);
     CheckFileName(OutputFileName, FormatNamePBM);

     /* Initialize output image files and data strucutres */
     CopyImageHeader(InputImage, &(OutputImage[i]));
     OutputImage[i].FileType = PBM;
     OutputFile[i] = OpenFile(OutputFileName, OUTPUT, Value(OverWrite));
     InitializeImage(OutputFile[i], &(OutputImage[i]), OUTPUT, WINDOWBUFFER, 1);
     }

  if(Value(QuietLevel)>=2) printf("--- Preparations ok.\n");

  /* Prepare quantized image */
  if(Value(OutputQuantizedImage)==YES)
     {
     PickFileName(PlainName, OutputFileName);
     sprintf(OutputFileName, "%s-Q", PlainName);
     CheckFileName(OutputFileName, FormatNamePPM);
     CopyImageHeader(InputImage, &QuantizedImage);
     QuantizedFile = OpenFile(OutputFileName, OUTPUT, Value(OverWrite));
     InitializeImage(QuantizedFile, &QuantizedImage, OUTPUT, WINDOWBUFFER, 1);
     }

  /* Analyze the input image line by line */
  RewindImage(InputImage);
  for( y=1; y<=InputImage->ImageSizeY; y++ )
     {
     ScrollBuffer(InputImage);
     /* Perform color separation for the current line of input image */
     for( x=1; x<=InputImage->ImageSizeX; x++ )
        {
		for(i=1; i<=BookSize(C); i++)
		   {
           PutBufferGray(&(OutputImage[i]), x, 0, 0);
		   }
		color = GetBufferRGBPixel(InputImage, x, 0);
        /* printf("(%03i,%03i,%03i)  ",color.Red, color.Green, color.Blue); */
        index = FindNearestColor(C, color);
        PutBufferGray(&(OutputImage[index+1]), x, 0, 1);
        if(Value(OutputQuantizedImage)==YES)
           {
           color2.Red   = VectorScalar(C, index, 0);
           color2.Green = VectorScalar(C, index, 1);
           color2.Blue  = VectorScalar(C, index, 2);
           PutBufferRGBPixel(&QuantizedImage, x, 0, color2);
           }
        }

     /* Output the current line for every output image */
     for(i=1; i<=BookSize(C); i++)
        {
        WriteCurrentLine(&(OutputImage[i]));
        }
     if(Value(OutputQuantizedImage)==YES)
        {
        WriteCurrentLine(&QuantizedImage);
        }

     PrintProgress(y, InputImage->ImageSizeY);
	 }

  if(Value(QuietLevel)>=2) printf("--- Separation ok.\n");

  for(i=1; i<=BookSize(C); i++)
     {
     FreeImage(&(OutputImage[i]));
     fclose(OutputFile[i]);
     }
  if(Value(OutputQuantizedImage)==YES)
     {
     FreeImage(&QuantizedImage);
     fclose(QuantizedFile);
     }
  if(Value(QuietLevel)>=2) printf("--- Deallocations ok.\n");
}


/*-------------------------------------------------------------------*/


static void  ProcessImage(IMAGE* InputImage, char* OutputFileName)
{
  TRAININGSET	TS;
  CODEBOOK      C;
  PARTITIONING	P;
  BINTREE		Tree;
  #define		MAX_CLUSTERS   256

  InitBintree(&Tree, CompareNodes);
  ReadPixelsToBinaryTree(InputImage, &Tree);
  InitializeTrainingSet(&TS, BintreeSize(&Tree), InputImage->MaxValue);
  ConvertTreeToTrainingSet(&TS, &Tree);
  CreateNewCodebook(&C, MAX_CLUSTERS, &TS);
  CreateNewPartitioning(&P, &TS, MAX_CLUSTERS);
  GenerateCodebook(&TS, &C, &P);
  PerformColorSeparation(InputImage, OutputFileName, &C);

  FreeBintree(&Tree);
  FreeCodebook(&TS);
  FreeCodebook(&C);
  FreePartitioning(&P);
}


/*===========================  M A I N  =============================*/


void main(int argc, char* argv[])
{
  char                InputFileName[64] = { '\0' };
  char                OutputFileName[64] = { '\0' };
  FILE*               InputFile;
  IMAGE               InputImage;
  long                watch;
  ParameterInfo paraminfo[2] = { { InputFileName, "PPM", 0, INFILE  },
                                 { OutputFileName, "",   1, OUTFILE } };

  ParseParameters(argc, argv, 2, paraminfo);
  InputFile  = OpenFile(InputFileName,  INPUT,  NO);
  InitializeImage(InputFile,  &InputImage,	INPUT,	WINDOWBUFFER, 1);
  if(strlen(OutputFileName)==0) PickFileName(InputFileName, OutputFileName);
  PrintOperatingInfo(InputFileName, OutputFileName, &InputImage);
  SetWatch(&watch);

  ProcessImage(&InputImage, OutputFileName);

  if(Value(QuietLevel)>=2) printf("Back in main program.\n");

  PrintTime(watch);
  FreeImage(&InputImage);
  fclose(InputFile);
  if(Value(QuietLevel)>=2) printf("Before memory check.\n");
  checkmemory();
  if(Value(QuietLevel)>=2) printf("After memory check.\n");
}

